"4g-with is a REPL command that allows calling commands in a loop while changing only args for it."

from pathlib import Path


import subprocess, re
import sys, tarfile, stat 
import hashlib, time


import logging, os, sys


# Diff Status Codes
SAME = 0
DIFFERENT = 1
TROUBLE = 2


def execute_command(command):
    """
        Execute a command and return output, errors and status code.  
    """
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = process.communicate()
    # Output, Errors, Exit/Status Code
    return {
        "output": stdout.decode().rstrip(),
        "errors": stderr.decode().rstrip(),
        "status": process.returncode,
        "command": command
        }

def run_command(command):
    """
        Execute a command and return output, errors and status code.  
    """
    started = time.time()
    result = execute_command(command)
    ended = time.time()
    
    result["time"] = ended-started

    return result

def read_command(command, errors=[""], status=[0]):
    """Run Command and return just output.
    
    By default it expects no errors.
    
    if errors and status params are not None, values will get asserted for testing.
    
    If errors and/or status in params are set None, they will get omitted.
    
    It is allowed to have multiple acceptable statuses. e.g. diff returns 0 and 1 and it is fine
    
    @param: status List of Acceptable Exit Codes.    
    """
    result = execute_command(command)
    
    command_info = f"""Command: {result['command']}
Status Code: {result['status']}
Errors:    
{result['errors']}
[End Of Errors]
"""
    
    # Test errors according to preferences set in params
    if status is not None:
        assert result["status"] in status, command_info
    
    if errors is not None:
        assert result["errors"] in errors, command_info

    assert "output" in result, command_info
    
    # We only need output
    return result["output"]


home = Path.home()

forgyWithCache = home / ".cache/4g/with"
forgyWithHistory = forgyWithCache / "history"
withLog = forgyWithCache / "with.log"

def mkdirs(path):
    if not os.path.exists(forgyWithCache):
        os.makedirs(str(forgyWithCache))

mkdirs(str(forgyWithCache))
mkdirs(str(forgyWithHistory))


# Configure logging

logging.basicConfig(filename=withLog, level=logging.DEBUG)

# Log messages


def inputline(
        color = "Yellow",
        prompt = "% ",
        history = "test",
        readline = "rlwrap",
        args = sys.argv[1:]
    ):
    """Wrapper around readline tool (the rlwrap, a wrapper around gnu-readline in C)
    
    i can use any readline library when i figure out how to do it.
    Why so many wrappers? First make it work! Then make it beautiful!"""
    history = history if history else " ".join(args)
    history = forgyWithHistory / f"{history}.history"
    return read_command(f"{readline} -p{color} -S '{prompt}' -H '{history}' -o cat")
    
def with_repl(*args, **kwargs):
    "main function for 4g-with"
    while True:
        cmd = inputline(*args, **kwargs)
        print(cmd)
        if cmd == "exit":
            logging.debug(f"4g-with: exiting command: {args}")
            break
        elif not cmd:
            logging.debug(f'4g-with: do nothing in this command: {args}')
            # Do nothing if cmd is empty
            pass
        
        logging.debug(f'4g-with: command: {cmd}; {args}')



import logging
import sys
import textwrap
import optparse

class ParamParser:
    """Wrapper around optparse.
    Hybrid Param Parser.
    """
    def _new_parser(self):
        return optparse.OptionParser(epilog=textwrap.dedent("""\
    This is a test command.
    
    Report bugs to https://codeberg.org/tfed/4g/issues."""))

    def __init__(self, args=sys.argv[1:], commands={}):
        self.command = "help"
        
        ## Raw Arguments
        self.args = {
            # All original Arguments
            'original': args,
            # Global Arguments (Options Mostly)
            # default: = all args, otherwise only up to first keyword
            'master': args,
            # Command's Options with Arguments 
            'command': []
        }
        
        ## Options
        self.options = {
            'master': {},
            'command': {}
        }

        self.commands = commands

        ## Parsers
        # Global and Command Parsers
        self.parsers = {'master': self._new_parser()}
        # the rest of parsers        
        for cmd in self.commands:
            self.parsers[cmd] = self._new_parser()

        self.init()

    def get_flag_safely(self, namespace, key):
        try:
            return self.options[namespace][key]
        except():
            return 0
        
    def init(self):
        # Split arguments in master and command's areas 
        self.splitArgs()

        # Parse args and options
        self.parseArgs("master")
        self.parseArgs()
        
        # Allow skipping areas while processing.
        self.args["all"] = self.args["master"] + self.args["command"]
        self.options["all"] = self.options["master"] | self.options["command"]
        print()
        print()
        # Fix global flag counters
        for key in self.options["command"]:
            print("master options: ", key, [i for i in self.options["master"]])
            print("command options: ", key, [i for i in self.options["command"]])
            if key in self.options["master"]:
                print(f'{key} is a common flag')

                # quick aliases to the parser's configured options
                global_options = self.commands["master"]["options"]
                command_options = self.commands[self.command]["options"]
                # for i in command_options:
                #     print(i)
                # Make sure that we deal with flags only
                if key in global_options or key in command_options:
                    print(f"{key} flag is an option")
                    continue
                print(f"{key} flag")
                
                # quick aliases to current parsed options
                global_value = self.options["master"][key]
                value = self.options["command"][key]
                
                # Clear count if -no or -non was used in command namespace
                if False in [global_value, value]:
                    global_value = 0
                
                # Convert None and False values to 0 while getting counts of --verboses
                values = [0 if isinstance(value, type(None)) else value for value in [global_value, value]]
                print(values)
                # Update summarized result
                result = sum(values)
                self.options["command"][key] = result
                self.options["master"][key] = result
                self.options["all"][key] = result

        print("Arguments:")
        for key in self.options["master"]:
            print(f"{key} args: ")
        print("Arguments:")
        for key, value in self.args.items():
            print(f"{key} args: ", value)
        print("Options:")
        for key, value in self.options.items():
            print(f"{key} options: ", value)
        print()
    
    def _splitArgsByKeyword(self, keyword):
        if keyword in self.args["original"]:
            self.command = keyword
            index = self.args["original"].index(keyword)
            # split arguments by command so we can parse before args and after args.
            # End params are linked to the command, when params before are more global.
            self.args["master"] = self.args["original"][:index]
            self.args["command"] = self.args["original"][index+1:]
            # Merge options, but command's options will be marked as command's sub
            # vars(parse(self._args[1]));    {command}.{option} * vars(parse(self._args[2]))

    def splitArgs(self):
        # Usage: 4g <options> <command> <command.options> <args>
        for cmd in self.commands:
            if cmd in sys.argv[1:] and cmd != "master":
                self._splitArgsByKeyword(cmd)
                break
        else:
            # Otherwise if there is no cmd inside we can just drop help msg
            # And exit with help error
            sys.exit(2)

    def add_flag(self, parser, dest, *args, **kwargs):
        parser.add_option(f'--{dest}', *args, dest=dest, action='count', **kwargs)

        # Support --no- with --non- key prefixes to set keys False
        parser.add_option(f"--no-{dest}", dest=dest, action='store_false')
        parser.add_option(f"--non-{dest}", dest=dest, action='store_false')

    def add_flags(self, parser, command):
        if "flags" not in self.commands[command]:
            self.commands[command]["flags"] = {}

        flags = self.commands[command]["flags"]

        for dest, act in flags.items():
            print(f'{command}: {dest}: {act}')
            (args, kwargs) = act['args'], act['kwargs']
            self.add_flag(parser, dest, *args, **kwargs)

        ## Warning: Recursion:
        # We want to allow global arguments into command namespace as well
        if command != "master":
            self.add_flags(parser, "master")

    def add_options(self, parser, command):
        if "options" not in self.commands[command]:
            self.commands[command]["options"] = {}

        options = self.commands[command]["options"]

        for dest, opt in options.items():
            print(f'{command}: {dest}: {opt}')
            (args, kwargs) = opt['args'], opt['kwargs']
            parser.add_option(f"--{dest}", *args, **kwargs)

        ## Warning: Recursion:
        # Allow master arguments working inside command's namespace
        if command != "master":
            self.add_options(parser, "master")

    def parseArgs(self, command=None):
        if not command:
            command = self.command
        namespace = "master" if command == "master" else "command"
        
        if namespace == "command":
            assert self.options[namespace] == {}

        parser = self.parsers[command]
        parser.prog = f"4g {command}"
        
        self.add_options(parser, command)
        self.add_flags(parser, command)
        match command:
            case "master":
                # Only master's progname is an exception 
                parser.prog = "4g"
            case "help":
                parser.print_help()


        opts, urls = parser.parse_args(self.args[namespace])
        opts = vars(opts)
        
        self.options[namespace] |= opts
        self.args[namespace] = urls


"""
fourGet is script that works with hashes of git projects.
You can create 'aliases' to reuse projects under other pseudonyms or hashes.
When author changed his nickname of renamed project it can be noted here as well.
When we now about name changes, git project can be easily reused.
Alias and Pseudonyms will be stored at ~/.local/4g/aliases.db sqlite database.
"""

import hashlib, os, sys

from pathlib import Path

home = Path.home()
fourHome = home / ".local/share/4g/home"
# DB with all aliases to reuse git sources
fourAliases = fourHome / "aliases.db"
# Git Sources or local git projects
fourSources = fourHome / "src"
projects = home / "Projects" / "4g"

def mkdirs(dir):
    try:
        os.makedirs(str(dir))
    except:
        pass

mkdirs(fourSources)
mkdirs(projects)

print(str(fourHome))

def hashedString(s, hashTool="sha256"):
    h = hashlib.new(hashTool)
    h.update(s.encode()) # give a encoded string. Makes the String to the Hash 
    return h.hexdigest() # Return the Hash

def hashForgySource(url, hashTool="md5"):
    urlHash = hashedString(url, hashTool)
    gitName = url.split("/")[-1]
    return f"{gitName}@{urlHash}"




"""
fourGive is script that git clones/pulls git projects.
It allows you to move projects in a special directory.
You can create 'aliases' to reuse projects under other pseudonyms or links.
It uses fourGet to redirect aliases to actual git projects.
"""

### Import libs with weird names 
import sys, os
import subprocess, re
import sys, tarfile, stat 
import hashlib, time

import os, pathlib
import arrow, time
from pathlib import Path

from datetime import datetime, timedelta
import pytz
from relative_datetime import DateTimeUtils

from forgy import fourGet


def cmd(command):
    with subprocess.Popen(command, shell=True) as process:
        process.communicate()

def execute_command(command):
    """
        Execute a command and return output, errors and status code.
        Note: Copied from transpkg.diffTools, when editing sync with the original.
    """
    with subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=True) as process:
        stdout, stderr = process.communicate()
        # Output, Errors, Exit/Status Code
        return {
            "output": stdout.decode().rstrip(),
            "errors": stderr.decode().rstrip(),
            "status": process.returncode,
            "command": command
            }
        

def listFiles(path):
    tagged_root_dir = pathlib.Path(path)
    return tagged_root_dir.resolve().glob('**/*')

def sortFilesByTime(path):
    return sorted(listFiles(path), key=os.path.getctime)

def findNewestFile(path):
    result = max(listFiles(path), key=os.path.getctime)
    return (result, os.path.getctime(result))


def findOldestFile(path):
    result = min(listFiles(path), key=os.path.getctime)
    return (result, os.path.getctime(result))


def pull(url, days=-7, hours=+5):
    gitPath = fourGet.fourSources / fourGet.hashForgySource(url)
    path = os.path.realpath(gitPath.absolute())
    gitHome = Path(gitPath / ".git")
    
    print(arrow.now() - timedelta(days=1))
    critical_time = arrow.now().shift(hours=hours).shift(days=days)
    print(critical_time.int_timestamp)
    time_at_hand = time.gmtime()
    print("  -==================tm mon: ", critical_time)
    if os.path.isdir(path):
        print(path)
        item_time = arrow.get(gitHome.stat().st_mtime)
        time_epoch = time.gmtime(item_time.int_timestamp)
        
        print("It's been %d years, %d months, %d days and %d hours, %d  minutes, %d seconds ago" % (time_at_hand.tm_year - time_epoch.tm_year, time_at_hand.tm_mon - time_epoch.tm_mon , time_at_hand.tm_mday - time_epoch.tm_mday, time_at_hand.tm_hour - time_epoch.tm_hour, time_at_hand.tm_min - time_epoch.tm_min, time_at_hand.tm_sec - time_epoch.tm_sec))
        print(item_time.int_timestamp)
        if item_time < critical_time :
            cmd(f"cd '{path}'; git pull")
        else:
            print("4g: pull: Git Project was already refreshed this week.")
    else:
        # only if not fresh:
        cmd(f"git clone '{url}' '{gitPath}'")



def install():
    """Install/Update anom's main projects.
    Do not use this daily."""
    # TODO: prevent refreshing git sources on daily bases.
    # To force frequent updates use --force-pull key
    pull("https://codeberg.org/anom/transpkg")
    pull("https://codeberg.org/anom/4g")
    # pull("https://codeberg.org/anom/mfs")
    # pull("https://codeberg.org/anom/upen")
    # pull("https://github.com/package-url/packageurl-python")

